(*
* Copyright (C) 2003 Universitaet Koblenz-Landau
* 
* This file is part of KRHyper.
* 
* KRHyper is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
* 
* KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
* details.
* 
* You should have received a copy of the GNU General Public License along with
* KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA 02111-1307 USA
* 
* Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>.
*)


type env =
    { symbol_table: Term.symbol_table;
      predicate_table: Predicate.predicate_table;
      negative_units: NegativeUnits.store;
      mutable rules: Term.term list;
      cfg: Cfg.cfg;
      mutable shared_input_table: SharedInput.table;
      mutable current_result:Tableau.result
    }

let make_env cfg =
  let st = Term.make_symbol_table () 
  and pt = Predicate.make_predicate_table ()
  and nu = NegativeUnits.make_store () in
  Builtin.init_builtins st;
  Evaluable.init_evaluables st;
  {symbol_table=st;
   predicate_table=pt;
   negative_units=nu;
   rules=[];
   cfg=cfg;
   shared_input_table=SharedInput.null_table;
   current_result=Tableau.initial_result cfg pt st nu
 }

let get_cfg env =
  env.cfg

let get_symbol_table env =
  env.symbol_table

let get_predicate_table env =
  env.predicate_table

let get_rules env =
  env.rules

let get_negative_units env =
  env.negative_units

type statement_type =
    Meta
  | Unit
  | Rule
  | Disjunction
  | False

let classify_statement_functor env sym =
  let st = env.symbol_table in
  if sym == Term.get_meta_functor_symbol st then
    Meta
  else if sym == Term.get_rule_symbol st then
    Rule
  else if sym == Term.get_disjunction_symbol st then
    Disjunction
  else if sym == Term.get_false_symbol st then
    False
  else
    Unit

let ensure_ready_to_assert env symbol =
  let pred = Predicate.find_predicate env.predicate_table symbol in
  Predicate.ensure_ready_to_assert pred;
  pred

let ensure_ready_to_run env symbol =
  let pred = Predicate.find_predicate env.predicate_table symbol in
  Predicate.ensure_ready_to_run pred;
  pred

let lookup_predicate env name arity =
  Predicate.PredicateTable.find 
    env.predicate_table
    (Term.find_symbol env.symbol_table name arity)

let abolish_predicate env name arity =
  Predicate.abolish
    env.predicate_table
    (Term.find_symbol env.symbol_table name arity);
  NegativeUnits.clear env.negative_units
    
let find_predicate env name arity =
  Predicate.find_predicate
    env.predicate_table
    (Term.find_symbol env.symbol_table name arity)

let add_rule env clause =
  env.rules <- clause :: env.rules

let add_disjunction env clause =
  env.rules <-
    (Term.Compound
       (Term.make_compound
	  (Term.get_rule_symbol env.symbol_table)
	  [|clause;
	    (Term.Constant (Term.get_true_symbol env.symbol_table))|])) ::
    env.rules

let add_false env =
  env.rules <-
    (Term.Compound
       (Term.make_compound
	  (Term.get_rule_symbol env.symbol_table)
	  [|(Term.Constant (Term.get_false_symbol env.symbol_table));
	    (Term.Constant (Term.get_true_symbol env.symbol_table))|])) ::
    env.rules


let clear_rules env =
  env.rules <- [];
  NegativeUnits.clear env.negative_units

let begin_shared_input env =
  if env.shared_input_table == SharedInput.null_table then
    env.shared_input_table <- SharedInput.make_table ()
  else
    ()

let end_shared_input env =
  env.shared_input_table <- SharedInput.null_table

let unit_for_env env term =
  if env.shared_input_table == SharedInput.null_table then
    term
  else
    SharedInput.share_subterms env.shared_input_table term

let get_current_result env =
    env.current_result

let set_current_result env model =
  env.current_result <- model
