(*
* Copyright (C) 2003 Universitaet Koblenz-Landau
* 
* This file is part of KRHyper.
* 
* KRHyper is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
* 
* KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
* details.
* 
* You should have received a copy of the GNU General Public License along with
* KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA 02111-1307 USA
* 
* Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>.
*)

type fact =
    InputFact of Term.term
  | DerivedFact of Term.term	
  | DerivedFactWithProof of Term.term * proof
  | FactWithUnknownProof
  | NullFact
      
and proof =
    BuiltinProof
  | InputFactProof
  | Proof of Term.term * fact array
  | DisjProof of int * Term.term * fact array * fact list
  | NegProof of Term.term * fact array * fact
  | UnknownProof
	
let get_term fact = 
  match fact with
    InputFact t -> t
  | DerivedFact t -> t
  | DerivedFactWithProof (t, _) -> t
  | FactWithUnknownProof -> Term.Unbound
  | NullFact -> Term.Unbound

let get_proof fact =
  match fact with
    InputFact _ -> InputFactProof
  | DerivedFact _ -> BuiltinProof
  | DerivedFactWithProof (_, p) -> p
  | FactWithUnknownProof -> UnknownProof
  | NullFact -> BuiltinProof

type entry = fact      

let null_fact = NullFact

let fact_with_unknown_proof = FactWithUnknownProof
    
let entry_print fact =
  PrologPrint.print (get_term fact)

let make_input_fact term = InputFact term

let make_subproofs n =
  Array.make n NullFact

let make_derived_fact term =
  DerivedFact term

let make_derived_fact_with_proof term proof =
  DerivedFactWithProof (term, proof)

let make_fact_like fact term =
  match fact with
    InputFact _ -> InputFact term
  | DerivedFact _ -> DerivedFact term
  | DerivedFactWithProof (_, proof) -> DerivedFactWithProof (term, proof)
  | FactWithUnknownProof -> FactWithUnknownProof
  | NullFact -> NullFact
	    
(****************************************************************************)

let make_ref_proofterm st nodenum =
  Convert.array_to_compound
    st
    "ref"
    [|Convert.string_to_term st ("node"^string_of_int nodenum)|]

let make_disj_ref_proofterm st nodenum =
  Convert.array_to_compound
    st
    "ref"
    [|Convert.string_to_term st ("disj"^string_of_int nodenum)|]
  
let make_input_fact_proofterm st nodenum term =
  Convert.array_to_compound
    st
    "node"
    [|Convert.string_to_term st ("node"^string_of_int nodenum);
      term;
      Convert.string_to_term st "input_fact";
      Convert.string_to_term st "[]";
      Convert.string_to_term st "[]"|]

let make_builtin_fact_proofterm st nodenum term =
  Convert.array_to_compound
    st
    "node"
    [|Convert.string_to_term st ("node"^string_of_int nodenum);
      term;
      Convert.string_to_term st "builtin";
      Convert.string_to_term st "[]";
      Convert.string_to_term st "[]"|]

let make_unknown_fact_proofterm st nodenum term =
  Convert.array_to_compound
    st
    "node"
    [|Convert.string_to_term st ("node"^string_of_int nodenum);
      term;
      Convert.string_to_term st "unknown";
      Convert.string_to_term st "[]";
      Convert.string_to_term st "[]"|]
    
let make_hyper_proofterm st nodenum term clause subproofs neg_subproofs =
  Convert.array_to_compound
    st
    "node"
    [|Convert.string_to_term st ("node"^string_of_int nodenum);
      term;
      Convert.array_to_compound st "hyper" [|clause|];
      Convert.list_of_term_to_term st (Array.to_list subproofs);
      Convert.list_of_term_to_term st neg_subproofs|]
    
let make_disj_proofterm st id clause subproofs neg_subproofs =
  Convert.array_to_compound
    st
    "split"
    [|Convert.string_to_term st ("disj"^string_of_int id);
      Convert.array_to_compound st "hyper" [|clause|];
      Convert.list_of_term_to_term st (Array.to_list subproofs);
      Convert.list_of_term_to_term st neg_subproofs|]

let make_branch_proofterm st nodenum term subproof =
  Convert.array_to_compound
    st
    "branch"
    [|Convert.string_to_term st ("node"^string_of_int nodenum);
      term;
      subproof|]

(* hmmm - what should be in the table is the "proof_id"
   - shared for diff goals in case of disjunction!
   - issues: space & explicitly different handling for input facts
   - perhaps just handle disjunctions specially
   - units: identify by their fact (including particular ancestors)
   - disjunction: identify by the proof term
*)

module Table = Hashtbl.Make(
  struct
    type t = fact
    let equal = (=) (* ### Fact-identity: = (and not ==) *)
    let hash = Hashtbl.hash
  end)

module DisjTable = Hashtbl.Make(
  struct
    type t = int
    let equal = (=)
    let hash = Hashtbl.hash
  end)

type sign = Pos | Neg

let negate st term =
  Convert.array_to_compound st "~" [|term|]

let rec proof_to_term_1 st sign table dtable node fact =
  try 
    let refnum = Table.find table fact in
    make_ref_proofterm st refnum
  with Not_found ->
    node := !node + 1;
    Table.add table fact !node;
    proof_to_term_2 st table dtable node 
      (match sign with Pos -> get_term fact | Neg -> negate st (get_term fact))
      (get_proof fact)

and proof_to_term_2 st table dtable node goal_term proof =
  match proof with
    BuiltinProof ->
      make_builtin_fact_proofterm st !node goal_term
  | UnknownProof ->
      make_unknown_fact_proofterm st !node goal_term
  | InputFactProof ->
      make_input_fact_proofterm st !node goal_term
  | Proof (clause, subfacts) ->
      let nodeval = !node in
      let subproofs =
	Array.map
	  (fun subfact -> proof_to_term_1 st Pos table dtable node subfact)
	  subfacts in
      make_hyper_proofterm st nodeval goal_term clause subproofs []
  | NegProof (clause, subfacts, negfact) ->
      let nodeval = !node in
      let subproofs =
	Array.map
	  (fun subfact -> proof_to_term_1 st Pos table dtable node subfact)
	  subfacts in
      let negproof = proof_to_term_1 st Neg table dtable node negfact in
      make_hyper_proofterm st nodeval goal_term clause subproofs [negproof]
  | DisjProof (id, clause, subfacts, negfacts) ->
      let nodeval = !node in
      make_branch_proofterm
	st 
	nodeval
	goal_term
	(try 
	  let dnum = DisjTable.find dtable id in
	  make_disj_ref_proofterm st dnum
	with Not_found ->
	  DisjTable.add dtable id id;
	  let subproofs =
	    Array.map
	      (fun subfact -> proof_to_term_1 st Pos table dtable node subfact)
	      subfacts in
	  let negproofs =
	    List.map
	      (fun negfact -> proof_to_term_1 st Neg table dtable node negfact)
	      negfacts in
	  make_disj_proofterm st id clause subproofs negproofs)
	  
let proof_term_for_fact symbol_table fact =
  let table = Table.create 64 
  and dtable = DisjTable.create 64 
  and node = ref 0 in
  proof_to_term_1 symbol_table Pos table dtable node fact

type l_proof = proof list

type proof_term_aux =
    {table:int Table.t;
     dtable:int DisjTable.t;
     node:int ref}

let make_proof_term_aux () =
  {table=Table.create 64;
   dtable=DisjTable.create 64;
   node=ref 0}
    
let proof_term_for_l_proof symbol_table proofs =
  let aux = make_proof_term_aux () in
  let table = aux.table
  and dtable = aux.dtable
  and node = aux.node in
  match proofs with
    [proof] ->
      proof_to_term_2 
	symbol_table
	table
	dtable
	node
	(Term.Constant (Term.get_false_symbol symbol_table))
	proof
  | _ ->
      let proof_terms =
	(List.map
	   (fun proof ->
	     node := !node + 1;
	     proof_to_term_2 
	       symbol_table
	       table
	       dtable
	       node
	       (Term.Constant (Term.get_false_symbol symbol_table))
	       proof)
	   proofs) in
      Convert.array_to_compound
	symbol_table
	"node"
	[|Convert.string_to_term symbol_table "node_goal";
	  Convert.string_to_term symbol_table "goal";
	  Convert.string_to_term symbol_table "all_closed";
	  Convert.list_of_term_to_term symbol_table proof_terms;
	  Convert.list_of_term_to_term symbol_table []|]

let proof_term_for_r_proof symbol_table aux proof =
  let table = aux.table
  and dtable = aux.dtable
  and node = aux.node in
  node := !node + 1;
  proof_to_term_2 
    symbol_table
    table
    dtable
    node
    (Term.Constant (Term.get_false_symbol symbol_table))
    proof
