(*
* Copyright (C) 2003 Universitaet Koblenz-Landau
* 
* This file is part of KRHyper.
* 
* KRHyper is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
* 
* KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
* details.
* 
* You should have received a copy of the GNU General Public License along with
* KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA 02111-1307 USA
* 
* Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>.
*)

type sign =
    Pos
  | StratifiedNegAsFailure
  | StratifiedNegAsFailureNotSubsumed
  | Findall
  | Neg2

let lit_sign_and_atom st term =
  let f = Convert.term_functor_symbol term in
  if f == (Term.get_stratified_neg_as_failure_symbol st) ||
     f == (Term.get_stratified_neg_as_failure_alternate_symbol st) 
  then
    (StratifiedNegAsFailure, Convert.term_nth_arg term 0)
  else if f == (Term.get_stratified_neg_as_failure_not_subsumed_symbol st) then
    (StratifiedNegAsFailureNotSubsumed, Convert.term_nth_arg term 0)
  else if f == (Term.get_findall_symbol st) then
    (Findall, Convert.term_nth_arg term 1)
  else if f == (Term.get_neg2_symbol st) then
    (Neg2, Convert.term_nth_arg term 0)
  else
    (Pos, term)

(****************************************************************************)
(* Auxiliary *)
(****************************************************************************)

let renumber_variables term =
  let table = Array.make Term.max_vars (-1) in
  Term.renumber_variables_term table term;
  term


(****************************************************************************)
(* Body Literal *)
(****************************************************************************)

module BodyLiteral =
  struct


(****************************************************************************)
(* Optim Info:  Information about the optimization intended for humans. *)
(****************************************************************************)

    type optim_info =
	{ applicability_hint: Opt.applicability_value;
	  discrimination_hint: Opt.discrimination_value;
	  binding_pattern: Opt.binding_pattern;
	  extent_size: int }

    let make_optim_info
	applicability_hint
	discrimination_hint
	binding_pattern
	extent_size =
      { applicability_hint=applicability_hint;
	discrimination_hint=discrimination_hint;
	binding_pattern=binding_pattern;
	extent_size=extent_size }

    let make_default_optim_info () =
      make_optim_info 0.0 0.0 [||] 0 

(****************************************************************************)
(* BodyLiteral *)
(****************************************************************************)

    type literal =
	{mutable sign: sign;
	 mutable predicate: Predicate.predicate;
	 mutable atom: Term.term;
	 mutable term: Term.term;
	 mutable reader: Access.Reader.reader;
	 mutable foreign_context: Term.context;
	 mutable is_delta: bool;
	 mutable optim_info: optim_info;
	 mutable builtin: Term.p;
	 mutable subproof_pos: int;
	 mutable output_vars: int list}

    let get_sign lit = lit.sign

    let get_predicate lit = lit.predicate

    let get_atom lit = lit.atom

    let get_reader lit = lit.reader

    let get_foreign_context lit = lit.foreign_context

    let get_is_delta lit = lit.is_delta

    let get_optim_info lit = lit.optim_info

    let get_builtin lit = lit.builtin

    let get_subproof_pos lit = lit.subproof_pos

    let set_subproof_pos lit n = lit.subproof_pos <- n
	

(****************************************************************************)
(* "Simple" Negation - i.e. Stratified Negation as Failure *)
(****************************************************************************)

(* Construct a builtin function, that interns the reader. *)

    exception Neg_failed
	
    let make_simpleneg_pdeterministic predicate =
      let reader = Predicate.get_all_bound_reader predicate in
      Term.PDeterministic
	((fun term context foreign_context trail ->
	  let saved_trail = Term.save_trail trail in
	  try
	    Access.Reader.iter_unification_candidates    
	      (fun entry ->
		let trail1 =
		  Term.unify
		    term
		    context
		    (Fact.get_term entry)
		    foreign_context
		    trail in
		if Term.failed trail1 then
		  ()
		else
		  begin
		    ignore (Term.undo_bindings trail1 saved_trail);
		    raise Neg_failed
		  end)
	      reader
	      term
	      context;
	    trail
	  with Neg_failed ->
	    Term.fail),
	 let arity = (Term.symbol_arity (Predicate.get_symbol predicate)) in
	 (Array.make arity Opt.InputGround,
	  Array.make arity 1.0,
	  Opt.UnknownProof))

(****************************************************************************)
(* Stratified Negation as Failure - "Not Subsumed" Variation *)
(****************************************************************************)

    let make_subsneg_pdeterministic predicate =
      Term.PDeterministic
	((fun term context foreign_context trail ->
	  try
	    Access.Reader.iter_generalization_candidates 
	      (fun fact -> 
		if Term.subsumes_chk
		    (Fact.get_term fact)
		    (* ### This is not the best implementation, since
		       the term is copied here - to avoid this, we
		       would need a subsumes_chk that takes a context
		       produced by unify. Copying is done here
		       only for generalization candidates. *)
		    (Term.apply term context) 
		then
		  raise Neg_failed
		else
		  ())
	      (Predicate.get_all_bound_reader predicate)
	      term
	      context;
	    trail
	  with Neg_failed ->
	    Term.fail),
	 let arity = (Term.symbol_arity (Predicate.get_symbol predicate)) in
	 (Array.make arity Opt.OutputFree,
	  Array.make arity 1.0,
	  Opt.UnknownProof))

(****************************************************************************)
(* Findall *)
(****************************************************************************)

(* Like Prolog's findall(Absterm, Query, List) but with sorted output
   list. Query argument must be a single positive literal. Variables
   in Absterm may only appear in Absterm and Query, and not anywhere else in
   the clause. All variables in Query that are not in term must appear
   in another subgoal (i.e. bound at call time). 

   Actually a findall subgoal is (### currently ?) handled with a
   mode that is InputGround for all Query argument positions that
   contain a variable not in Absterm.
*)


let rec term_list_rm_duplicates list =
  match list with
    e1 :: e2 :: list1 ->
      if (StdOrder.term_compare
	    e1 Term.null_context e2 Term.null_context) = 0 then
	term_list_rm_duplicates (e2 :: list1)
      else
	e1 :: term_list_rm_duplicates (e2 :: list1)
  | _ -> list
	
let make_findall_function symbol_table reader absterm list =
  let dot = Term.find_symbol symbol_table "." 2
  and nil = Term.find_symbol symbol_table "[]" 0 in
  (fun query context foreign_context trail ->
    let results = ref [] in
    Access.Reader.iter_unification_candidates
      (fun entry ->
	let saved_trail = Term.save_trail trail in
	let trail1 =
	  Term.unify
	    query
	    context
	    (Fact.get_term entry)
	    foreign_context
	    trail in
	if Term.failed trail1 then
	  ()
	else
	  begin
	    Listutils.push
	      (renumber_variables (Term.apply absterm context))
	      results;
	    ignore (Term.undo_bindings trail1 saved_trail)
	  end)
      reader
      query
      context;
    let output_list =
      Convert.list_of_term_to_term_1
	dot nil
	(term_list_rm_duplicates
	   (List.sort
	      (fun e1 e2 -> 
		StdOrder.term_compare 
		  e1 Term.null_context e2 Term.null_context)
	      !results)) in
    Term.unify
      list
      context
      output_list
      context
      trail)

exception Found

let make_findall_pdeterministic symbol_table predicate query absterm list =
  let absvars = ref [] in
  Termutils.iter_vars
    (fun v -> Listutils.pushnew (Term.variable_number v) absvars)
    absterm;
  let arity = (Term.symbol_arity (Predicate.get_symbol predicate)) in
  let mode = Array.make arity Opt.OutputGround
  and bp = Array.make arity Opt.Free in
  if arity > 0 then
    match query with
      Term.Compound k ->
	let args = Term.compound_args k in
	for i = 0 to (arity - 1) do
	  try
	    Termutils.iter_vars
	      (fun v -> 
		if (not (List.mem (Term.variable_number v) !absvars)) then
		  raise Found
		else
		  ())
	      args.(i)
	  with Found ->
	    mode.(i) <- Opt.InputGround;
	    bp.(i) <- Opt.Bound
	done
    | _ ->
	raise (Failure "Bad findall arguments.")
  else
    ();
  let reader = Predicate.choose_best_reader predicate bp in
  Term.PDeterministic
    ((make_findall_function symbol_table reader absterm list),
       ( mode,
	 Array.make arity 1.0,
	 Opt.UnknownProof ))
    

(****************************************************************************)
(* BodyLiteral Constructors *)
(****************************************************************************)

    let aov_1 mode args vars =
      for i = 0 to (Array.length mode) - 1 do
	match mode.(i) with
	  Opt.OutputGround ->
	    Termutils.iter_vars
	      (fun v -> Listutils.pushnew 
		  (Term.variable_number v) vars)
	      args.(i)
	| Opt.InputGround | Opt.OutputFree ->
	    ()
      done

    let assign_output_vars lit term =
      let vars = ref [] in
      (match lit.sign with
	StratifiedNegAsFailure | StratifiedNegAsFailureNotSubsumed | Neg2 ->
	  ()
      | Findall ->
	  (* The 3rd arg is not contained in the atom. *)
	  Termutils.iter_vars
	    (fun v -> Listutils.pushnew (Term.variable_number v) vars)
	    (Convert.term_nth_arg lit.term 2)
      | Pos ->
	  match lit.atom with
	    Term.Compound compound ->
	      (match lit.builtin with
		Term.PDeterministic (_, (mode, _, _)) ->
		  aov_1 mode (Term.compound_args compound) vars
	      | Term.PUnbound ->
		  try
		    match 
		      List.find
			(fun decl -> 
			  match decl with Opt.ModeDecl _ -> true | _ -> false)
			(Predicate.get_declarations lit.predicate)
		    with Opt.ModeDecl mode ->
		      aov_1 mode (Term.compound_args compound) vars
		    | _ ->
			()
		  with Not_found ->
		    Termutils.iter_vars
		      (fun v -> 
			Listutils.pushnew (Term.variable_number v) vars)
		      lit.atom)
	  | _ ->
	      ());
      lit.output_vars <- !vars

    let make_literal predicate_table symbol_table term =
      let (sign, atom) = lit_sign_and_atom symbol_table term in
      let pred = (Predicate.find_predicate
		    predicate_table
		    (Convert.term_functor_symbol atom)) in
      Predicate.ensure_ready_to_run pred;
      let builtin =
	match sign with
	  StratifiedNegAsFailure ->
	    (make_simpleneg_pdeterministic pred)
	| StratifiedNegAsFailureNotSubsumed ->
	    (make_subsneg_pdeterministic pred)
	| Findall ->
	    (make_findall_pdeterministic
	       symbol_table
	       pred 
	       atom
	       (Convert.term_nth_arg term 0)
	       (Convert.term_nth_arg term 2))
	| _ ->
	    Term.symbol_predicate (Predicate.get_symbol pred) in
      let lit =
	{sign=sign;
	 predicate=pred;
	 term=term;
	 atom=atom;
	 is_delta=false;
	 builtin=builtin;
	 output_vars=[];
	 optim_info=make_default_optim_info (); (* Dummy init value. *)
	 reader=Access.Reader.null_reader (); (* Dummy init value. *)
	 foreign_context=Term.make_context 0; (* Dummy init value. *)
	 subproof_pos=(-1)} in (* Dummy init value. *)
      assign_output_vars lit term;
      lit
	  
    (** Returns a shallow copy of the given literal object, i.e.
	field values, including the atom term are shared. *)
    let make_literal_like literal =
      {sign=literal.sign;
       predicate=literal.predicate;
       atom=literal.atom;
       term=literal.term;
       is_delta=literal.is_delta;
       reader=literal.reader;
       foreign_context=literal.foreign_context;
       optim_info=literal.optim_info;
       builtin=literal.builtin;
       subproof_pos=literal.subproof_pos;
       output_vars=literal.output_vars}
	 
    let print_binding_pattern bp =
      for i = 0 to (Array.length bp) - 1 do
	match bp.(i) with
	  Opt.Bound ->
	    print_string "b"
	| Opt.Free ->
	    print_string "f"
      done

    let print_optim_info lit =
      let optim_info = lit.optim_info in
      Printf.printf "applic: %1.5f disc: %1.5f extent: %7d bpattern: "
	optim_info.applicability_hint
	optim_info.discrimination_hint
	optim_info.extent_size;
	print_binding_pattern optim_info.binding_pattern;
      print_endline ""
	
  end


(****************************************************************************)
(* HeadLiteral *)
(****************************************************************************)

module HeadLiteral =
  struct

    type literal =
	{mutable atom: Term.term;
	 mutable predicate: Predicate.predicate}

    let get_atom lit = lit.atom

    let get_predicate lit = lit.predicate

    let make_literal predicate_table term =
      let atom=term in
      let pred = (Predicate.find_predicate
		    predicate_table
		    (Convert.term_functor_symbol atom)) in
      Predicate.ensure_ready_to_run pred;
      {predicate=pred;
       atom=atom}
  end
  
    
(****************************************************************************)
(* Aux - Auxiliary datastructures passed to and kept by clauses. *)
(****************************************************************************)

module Aux =

  struct 

    type aux = {
	predicate_table: Predicate.predicate_table;
	symbol_table: Term.symbol_table;
	cfg: Cfg.cfg;
	negative_units_store: NegativeUnits.store;
	use_negative_units: bool;
	use_proof_terms: bool;
	mutable proof_id: int
      }

    let make_aux cfg predicate_table symbol_table neg_store  =
      { cfg=cfg;
	predicate_table=predicate_table;
	symbol_table=symbol_table;
	negative_units_store=neg_store;
	use_negative_units=Cfg.get_negative_units_flag cfg;
	use_proof_terms=Cfg.get_proof_terms_flag cfg;
	proof_id=0
      }

  end

(****************************************************************************)
(* Disjunction *)
(****************************************************************************)

module DisjunctiveHead =

  struct

    type head =
	{literals:HeadLiteral.literal list}

    let get_literals disjunction =
      disjunction.literals

    let make_head literals =
      {literals=literals}

  end

(****************************************************************************)
(* Clause *)
(****************************************************************************)

let id = ref 0

let gen_id () =
  id := !id + 1;
  "clause_"^string_of_int(!id)

let reset_gen_id () =
  id := 0
	  
type body =
    Body of BodyLiteral.literal list

type head =
    False
  | Unit of HeadLiteral.literal
  | Conjunction of HeadLiteral.literal list
  | Disjunction of DisjunctiveHead.head
	
type clause = {
    id: string;
    source: Term.term;
    mutable body: body;
    head: head;
    mutable news: News.news;
    mutable subproofs: Fact.fact array;
    cfg: Cfg.cfg;
    aux: Aux.aux
  }
    
let get_body clause = clause.body

let get_head clause = clause.head

let get_news clause = clause.news

let get_cfg clause = clause.cfg

let set_news clause news =
  clause.news <- news

let get_id clause = clause.id

let get_source clause = clause.source

let rec rm_constant symbol terms =
  match terms with
    [] -> []
  | term :: terms1 ->
      if Convert.term_functor_symbol term == symbol then
	rm_constant symbol terms1
      else
	term :: rm_constant symbol terms1
		 
let make_head predicate_table symbol_table term =
  let f = Convert.term_functor_symbol term in
  if f == Term.get_disjunction_symbol symbol_table then
    let lits =
      List.map 
	(fun lit -> HeadLiteral.make_literal predicate_table lit)
	(rm_constant
	   (Term.get_false_symbol symbol_table)
	   (Convert.term_sequence_to_list
	      (Term.get_disjunction_symbol symbol_table)
	      term)) in
    match lits with
      [] -> False
    | [lit] -> Unit lit
    | _ -> Disjunction (DisjunctiveHead.make_head lits)
  else if f == Term.get_conjunction_symbol symbol_table then
    let lits =
      List.map 
	(fun lit -> HeadLiteral.make_literal predicate_table lit)
	(Convert.term_sequence_to_list
	   (Term.get_conjunction_symbol symbol_table)
	   term) in
    Conjunction lits
  else if f == Term.get_false_symbol symbol_table then
    False
  else
    Unit (HeadLiteral.make_literal predicate_table term)

let make_body predicate_table symbol_table term =
  Body (List.map
	  (fun lit -> 
	    BodyLiteral.make_literal predicate_table symbol_table lit)
	  (rm_constant
	     (Term.get_true_symbol symbol_table)
	     (Convert.term_sequence_to_list
		(Term.get_conjunction_symbol symbol_table)
		term)))

let make_clause aux term =
  let cfg = aux.Aux.cfg
  and predicate_table = aux.Aux.predicate_table
  and symbol_table = aux.Aux.symbol_table in
  (* Assume term has :-/2 as main functor. *)
  let head = 
    make_head predicate_table symbol_table (Convert.term_nth_arg term 0)
  and body = 
    make_body predicate_table symbol_table (Convert.term_nth_arg term 1)
  and id = gen_id ()
  in
  {id=id;
   source=term;
   body=body;
   head=head;
   news=News.null_news;
   subproofs=[||];
   cfg=cfg;
   aux=aux
 }
    
let body_literals clause =
  match clause.body with
    Body lits -> lits
    
(** Returns a "shallow" copy of the given clause: the body has copies
   of the original literals, which however may share terms with
   the original ones, see [make_literal_like].
 *)
let make_clause_like clause =
  { id=clause.id;
    source=clause.source;
    news=clause.news;
    body=Body (List.map 
		 BodyLiteral.make_literal_like
		 (body_literals clause));
    head=clause.head;  (* ### may also make copies if needed. *)
    subproofs=Array.copy clause.subproofs;
    cfg=clause.cfg;
    aux=clause.aux 
  }
    
let clause_delta_clauses delta_predicates clause =
  let cs = ref []
  and i = ref 0 in
  List.iter 
    (fun lit ->
      let pred =lit.BodyLiteral.predicate in
      if List.memq pred delta_predicates then
	begin
	  Predicate.ensure_delta_ready_to_run pred;
	  let clause1 = make_clause_like clause in
	  (List.nth (body_literals clause1) !i).BodyLiteral.is_delta <- 
	    true;
	  cs := clause1 :: !cs
	end
      else
	();
      i := !i + 1)
    (body_literals clause);
  !cs
    
    
(****************************************************************************)
(* Subgoal Sorting. *)
(****************************************************************************)

(* For now this is greedy, i.e. no backtracking ("query planning")
   involved. 
*)
    
let lit_arity lit =
  Term.symbol_arity (Predicate.get_symbol lit.BodyLiteral.predicate)
    
let lit_nth_arg lit n =
  Convert.term_nth_arg lit.BodyLiteral.atom n
    
let mk_binding_pattern bound_vars lit =
  let arity = lit_arity lit in
  let bp = Array.make arity Opt.Free in
  for i = 0 to arity - 1 do
    let arg = (lit_nth_arg lit i) in
    if
      Termutils.is_nonvar arg ||
      List.mem (Termutils.varnum arg) bound_vars then
      bp.(i) <- Opt.Bound
    else
      ()
  done;
  bp
    
let permissible_by_modespec bound_vars modespec disc_hints lit =
  let arity = lit_arity lit in
  for i = 0 to arity - 1 do
    if modespec.(i) = Opt.InputGround then
      Termutils.iter_vars
	(fun v -> 
	  if not (List.mem (Term.variable_number v) bound_vars) then
	    raise Not_found
	  else
	    ())
	(lit_nth_arg lit i)
  done;
  let bp = (mk_binding_pattern bound_vars lit) in
  (1.0, 
   Opt.compute_discrimination_hint bp disc_hints,
   bp)
    
let permissible bound_vars lit =
  (* Returns (applicability_hint, discrimination_hint). Raises
     Not_found if the literal is not applicable at all with the
     given binding pattern. *)
  match lit.BodyLiteral.builtin with
    Term.PDeterministic (_, (mode, discs, _)) ->
      permissible_by_modespec bound_vars mode discs lit
  | Term.PUnbound ->
      let bp = mk_binding_pattern bound_vars lit
      and pred = lit.BodyLiteral.predicate in
      (Predicate.applicability pred bp,
       Opt.compute_discrimination_hint
	 bp (Predicate.get_discrimination_hints pred),
       bp)
	
let rec split_permissibles bound_vars lits y n =
  match lits with
    [] ->
      (y, n)
  | lit :: lits1 ->
      try
	let adb = permissible bound_vars lit in
	split_permissibles bound_vars lits1 ((lit, adb)::y) n
      with Not_found ->
	split_permissibles bound_vars lits1 y (lit::n)
	  

(****************************************************************************)

let compare_permissibles_by_to_end (l1, _) (l2, _) =
  if l1.BodyLiteral.sign = StratifiedNegAsFailureNotSubsumed then
    if l2.BodyLiteral.sign = StratifiedNegAsFailureNotSubsumed then
      0
    else
      1
  else if l2.BodyLiteral.sign = StratifiedNegAsFailureNotSubsumed then
    -1
  else
    0
      	  
let compare_permissibles_by_delta (l1, _) (l2, _) =
  if l1.BodyLiteral.is_delta then
    if l2.BodyLiteral.is_delta then
      0
    else
      -1
  else if l2.BodyLiteral.is_delta then
    1
  else
    0
      
let compare_permissibles_by_applicability_hint
    (_, (a1, _, _)) (_, (a2, _, _)) =
  compare a2 a1
    
let compare_permissibles_by_discrimination_hint 
    (_, (_, d1, _)) (_, (_, d2, _)) =
  compare d2 d1
    
let compare_permissibles_by_extent_size (l1, _) (l2, _) =
  (* ### 0 is preferred, but it effectively means "unknown", i.e.
     the effect is that IDB predicates are preferred. 
     Maybe change this. *)
  compare 
    (Predicate.extent_size l1.BodyLiteral.predicate)
    (Predicate.extent_size l2.BodyLiteral.predicate)
    
let compare_permissibles_somehow (l1, _) (l2, _) =
  (* Used to be independent on input ordering. *)
  compare 
    (l1.BodyLiteral.sign, l1.BodyLiteral.atom)
    (l2.BodyLiteral.sign, l2.BodyLiteral.atom)
    
let compare_permissibles lh1 lh2 =
  let r = compare_permissibles_by_to_end lh1 lh2 in
  if r <> 0 then r
  else
    let r = compare_permissibles_by_delta lh1 lh2 in
    if r <> 0 then r
    else
      let r = compare_permissibles_by_applicability_hint lh1 lh2 in
      if r <> 0 then r
      else
	let r = compare_permissibles_by_discrimination_hint lh1 lh2 in
	if r <> 0 then r
	else
	  let r = compare_permissibles_by_extent_size lh1 lh2 in
	  if r <> 0 then r
	  else
	    compare_permissibles_somehow lh1 lh2
	    
let choose_from_permissibles lhs =
  match List.sort compare_permissibles lhs with
    (l, (a, d, b)) :: rest ->
      (l, List.map fst rest, (a, d, b))
  | _ -> raise Not_found

let add_bound_vars bound_vars lit =
  List.append lit.BodyLiteral.output_vars bound_vars

let assign_optim_info lit (a, d, b) =
  lit.BodyLiteral.optim_info <-
    BodyLiteral.make_optim_info
      a d b (Predicate.extent_size lit.BodyLiteral.predicate)

(****************************************************************************)

(* Heuristics: if the greedy best next literal has application-hint and
   discrimation-hint of 0.0 - i.e. no index can be applied, the literal
   is chosen in a different way: that one which has the best successor
   literal if it would be the first is chosen.
*)

let is_poor_choice lit (applic_hint, discrim_hint, _) =
   (not lit.BodyLiteral.is_delta) &&
   applic_hint = 0.0 &&
   discrim_hint = 0.0

let rec map_split bef es  =
  match es with
    [] -> []
  | e::es1 ->
      (e, List.append bef es1) :: (map_split (e::bef) es1)

let choose_from_permissibles_by_second bound_vars lits =
  let lps = map_split [] lits in
  let lhs =
    List.map
      (fun (lit, lits1) ->
	let bound_vars_1 =
	  add_bound_vars bound_vars lit in
	let (next_permissibles, _) =
	  split_permissibles bound_vars_1 lits1 [] [] in
	match List.sort compare_permissibles next_permissibles with
	  (lit1, (a, d, b)) :: _ ->
	    (* Return this literal with values of the second one, obtained
	       by assuming that this literal is the first one. *)
	    (lit, (a, d, b))
	| _ ->
	    raise Not_found)
      lps in
  choose_from_permissibles lhs

(****************************************************************************)

let choose_permissible bound_vars lits =
  (* Throws Not_found. *)
  let (permissibles, rest1) =
    split_permissibles bound_vars lits [] [] in
  let (lit, rest2, h) =
    choose_from_permissibles permissibles in
  if is_poor_choice lit h && (List.length permissibles) > 1 then
    let (lit, rest2, h) =
      choose_from_permissibles_by_second
	bound_vars (List.map fst permissibles) in
    (lit, List.append rest1 rest2)
  else
    begin
      assign_optim_info lit h;
      (lit, List.append rest1 rest2)
    end


(****************************************************************************)

let rec select_delta deltaref lits =
  match lits with
    [] -> raise Not_found
  | lit :: lits1 ->
      if lit.BodyLiteral.is_delta then
	begin
	  deltaref := lit;
	  lits1
	end
      else lit :: select_delta deltaref lits1

let delta_to_front lits =
  match lits with
    [] -> []
  | lit :: _ ->
      let deltaref = ref lit in
	try
	  let rest = select_delta deltaref lits in
	  !deltaref :: rest
	with Not_found ->
	  lits


(****************************************************************************)
    
(** Raises Not_found if no permissible ordering can be found. *)

let rec sort_body_lits bound_vars lits =
  match lits with
    [] -> []
  | _ ->
      let (lit, lits1) = choose_permissible bound_vars lits in
      lit :: sort_body_lits 
	      (add_bound_vars bound_vars lit)
	      lits1
	  
(****************************************************************************)
	  
let rec assign_foreign_contexts n lits =
  match lits with
    [] ->
      ()
  | lit :: lits1 ->
      lit.BodyLiteral.foreign_context <- Term.make_context n;
      assign_foreign_contexts (n + 1) lits1
	
let rec assign_readers bound_vars lits =
  match lits with
    [] ->
      ()
  | lit :: lits1 ->
      begin
	match lit.BodyLiteral.builtin with
	  Term.PUnbound ->
	    let bp = mk_binding_pattern bound_vars lit
	    and pred = lit.BodyLiteral.predicate in
	    lit.BodyLiteral.reader <- 
	      Predicate.choose_best_reader
		(if lit.BodyLiteral.is_delta then
		  Predicate.get_delta_predicate pred
		else
		  pred)
		bp	  
	| _ ->
	    ()
      end;
      assign_readers
	(add_bound_vars bound_vars lit)
	lits1

let assign_subproof_stuff clause =
  if clause.aux.Aux.use_proof_terms then
    match clause.body with
      Body lits ->
	clause.subproofs <- Fact.make_subproofs (List.length lits);
	let i = ref 0 in
	List.iter
	  (fun lit -> BodyLiteral.set_subproof_pos lit !i; i := !i+1)
	  lits
  else
    ()
	    
let rec initialize_clause clause news =
  clause.news <- news;
  let bound_vars = [] in
  let lits = body_literals clause in
  let lits1 =
    match Cfg.get_body_sorting_method_param clause.cfg
    with
      0 ->
	(try
	  sort_body_lits bound_vars lits
	with Not_found ->
	  raise (Err.Error 
		   ("No permissible ordering for clause: "^
		    PrologPrint.print_to_string clause.source)))
    | 1 ->
	delta_to_front lits
    | 2 ->
	lits 
    | _ ->
	raise (Err.Error "Bad value for body_sorting_parameter.")
  in
  clause.body <- Body lits1;
  assign_readers bound_vars lits1;
  assign_foreign_contexts 1 lits1;
  assign_subproof_stuff clause


(** Restore the state of a clause, e.g. passed through a [BranchClosed]
   exception, such that it can be used for evaluation again. *)    
let refresh_clause clause =
  assign_foreign_contexts 1 (body_literals clause);
  assign_subproof_stuff clause


(****************************************************************************)
(* Runtime. *) 
(****************************************************************************)

exception BranchClosed of Fact.proof


(****************************************************************************)
(* Unit Heads *)

let check_new_unit clause fact =
  let aux = clause.aux in
  if aux.Aux.use_negative_units then
    try
      let neg_fact = NegativeUnits.lookup aux.Aux.negative_units_store fact in
      let proof =
	if aux.Aux.use_proof_terms then
	  Fact.NegProof (clause.source, 
			 (Array.copy clause.subproofs),
			 neg_fact)
	else
	  Fact.BuiltinProof in
      refresh_clause clause;
      raise (BranchClosed proof)
    with Not_found ->
      ()

let process_head_literal clause context lit =
  let fact =
    Fact.make_derived_fact
      (renumber_variables (Term.apply lit.HeadLiteral.atom context)) in
  check_new_unit clause fact;
  News.insert_unit
    clause.news 
    lit.HeadLiteral.predicate 
    fact
    
let process_head_literal_with_proof clause context lit =
  let fact =
    Fact.make_derived_fact_with_proof
      (renumber_variables (Term.apply lit.HeadLiteral.atom context))
      (Fact.Proof (clause.source,
		   (Array.copy clause.subproofs))) in
  check_new_unit clause fact;
  News.insert_unit
    clause.news 
    lit.HeadLiteral.predicate
    fact


(****************************************************************************)
(* Disjunctive Heads *)

let rec rcu_1 negunits nfs lits =
  match lits with
    [] -> []
  | lit :: lits1 ->
      try
	match lit with
	  (_, fact, _) ->
	    let nf = NegativeUnits.lookup negunits fact in

	    (* print_endline "RCU"; *)

	    nfs := nf :: !nfs;
	    rcu_1 negunits nfs lits1
      with Not_found ->
	lit :: rcu_1 negunits nfs lits1

let remove_complemented_units negunits lits =
  let nfs = ref [] in
  let lits1 = rcu_1 negunits nfs lits in
  (lits1, !nfs)
    
(** Assuming that sorting had put identical literals in succession. *)
let rec factorize_identical_literals lits =
  match lits with
    l1 :: l2 :: lits1 ->
      begin
	match l1, l2 with
	  (_, f1, _), (_, f2, _) ->
	    if Term.term_ident (Fact.get_term f1) (Fact.get_term f2) then
	      factorize_identical_literals (l2 :: lits1)
	    else
	      l1 :: factorize_identical_literals (l2 :: lits1)
      end
  | _ -> lits

let rec add_complement_splits lits =
  match lits with
    [] -> []
  | (p, f, w) :: lits1 ->
      (p, f, w, List.map (fun (_, f1, _) -> f1) lits1) ::
      add_complement_splits lits1

let process_disjunctive_head clause context head =
  let aux = clause.aux 
  and cfg = clause.cfg in
  let weighing_method =
    (* ### this dispatch is also in news.ml *)
    (match Cfg.get_weighing_method_param cfg with
      0 -> Weight.SymbolCount
    | 1 -> Weight.Depth
    | _ -> raise
	  (Err.Error "Bad value for term_weighing_method parameter.")) in
  let var_table = Array.make Term.max_vars (-1) in
  let lits1 =
    List.map
      (fun lit -> 
	let term = Term.apply lit.HeadLiteral.atom context in
	Term.renumber_variables_term var_table term;
	(lit.HeadLiteral.predicate,
	 Fact.make_derived_fact term,
	 (Weight.term_weight weighing_method term)))
      (DisjunctiveHead.get_literals head) in

  if Cfg.get_check_disjunctions_are_ground_flag cfg &&
    List.exists
      (fun (_, fact, _) -> 
	not (Termutils.is_ground (Fact.get_term fact)))
      lits1 then
    raise (Err.Error 
	     ("Nonground disjunct derived with clause: "^
	      (PrologPrint.print_to_string clause.source)));

  let (lits2, nfs) =
    if aux.Aux.use_negative_units then
      remove_complemented_units aux.Aux.negative_units_store lits1
    else
      (lits1, []) in
  (*
  print_int (List.length lits1); print_endline " lits1";
  print_int (List.length lits2);  print_endline " lits2";
  print_int (List.length nfs);  print_endline " nfs";
  *)

  let proof =
    if aux.Aux.use_proof_terms then
      begin
	aux.Aux.proof_id <- aux.Aux.proof_id + 1;
	Fact.DisjProof (aux.Aux.proof_id,
			clause.source,
			Array.copy clause.subproofs, 
			nfs)
      end
    else
      Fact.BuiltinProof in
  if lits2 = [] then
    begin
      refresh_clause clause;
      raise (BranchClosed proof)
    end
  else
    let lits2a =
      if aux.Aux.use_proof_terms then
	List.map
	  (fun (p,f,w) -> 
	    (p,
	     Fact.make_derived_fact_with_proof (Fact.get_term f) proof,
	     w))
	  lits2
      else
	lits2
    in
    let lits3 =
      if Cfg.get_sort_heads_flag cfg then
	List.stable_sort
	  (fun (_, _, w1) (_, _, w2) -> compare w1 w2)
	  lits2a
      else
	lits2a in
    let lits4 =
      factorize_identical_literals lits3 in
    match lits4 with
      [(pred, fact, _)] ->
	News.insert_unit clause.news pred fact
    | _ -> 
	let lits5 =
	  if aux.Aux.use_negative_units && 
	    not (Cfg.get_wp_semantics_flag cfg) then
	    add_complement_splits lits4
	  else
	    List.map (fun (p,t,w) -> (p,t,w,[])) lits4 in
	News.insert_disjunction clause.news lits5
	

(****************************************************************************)
(* Head Processing *)
      
let process_head clause context =
  match clause.head with
    False ->
      let proof = 
	if clause.aux.Aux.use_proof_terms then
	  Fact.Proof (clause.source, Array.copy clause.subproofs)
	else
	  Fact.BuiltinProof in
      refresh_clause clause;
      raise (BranchClosed proof)
  | Unit lit ->
      if clause.aux.Aux.use_proof_terms then
	process_head_literal_with_proof clause context lit
      else process_head_literal clause context lit
  | Conjunction lits ->
      if clause.aux.Aux.use_proof_terms then
	List.iter
	  (fun lit -> process_head_literal_with_proof clause context lit)
	  lits
      else
	List.iter
	  (fun lit -> process_head_literal clause context lit)
	  lits
  | Disjunction head ->
      (* TODO: ### Variant with proofs *)
      process_disjunctive_head clause context head
	

(****************************************************************************)
(* Clause Evaluation *)

let set_subproof clause lit entry =
  clause.subproofs.(BodyLiteral.get_subproof_pos(lit)) <- entry
      
let rec eval_clause clause =
  let context = Term.make_context 0
  and trail = Term.make_trail ()
  and literals = match clause.body with Body lits -> lits in
  eval_clause_1 clause context literals trail

and  eval_clause_1 clause context1 lits trail =
  match lits with
    [] -> 
      process_head clause context1
  |	lit :: lits1 ->
      match lit.BodyLiteral.builtin with
	Term.PUnbound ->
	  Access.Reader.iter_unification_candidates
	    (fun entry ->
	      let saved_trail=Term.save_trail trail in
	      let trail1 =
		Term.unify
		  lit.BodyLiteral.atom
		  context1 
		  (Fact.get_term entry)
		  lit.BodyLiteral.foreign_context
		  trail in
	      if Term.failed trail1 then
		()
	      else
		begin
		  if clause.aux.Aux.use_proof_terms then
		    set_subproof clause lit entry
		  else
		    ();
		  eval_clause_1 clause context1 lits1 trail1;
		  ignore (Term.undo_bindings trail1 saved_trail)
		end)
	    lit.BodyLiteral.reader
	    lit.BodyLiteral.atom
	    context1
      | Term.PDeterministic (f, (_, _, opt_proof)) ->
	  let saved_trail=Term.save_trail trail in
	  let trail1 =
	    f
	      lit.BodyLiteral.atom
	      context1
	      lit.BodyLiteral.foreign_context
	      trail
	  in
	  if Term.failed trail1 then
	    ()
	  else
	    begin
	      if clause.aux.Aux.use_proof_terms &&
		opt_proof = Opt.UnknownProof then
		set_subproof clause lit Fact.fact_with_unknown_proof
	      else
		();
	      eval_clause_1 clause context1 lits1 trail1;
	      ignore (Term.undo_bindings trail1 saved_trail)
	    end
