(*
* Copyright (C) 2003 Universitaet Koblenz-Landau
* 
* This file is part of KRHyper.
* 
* KRHyper is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
* 
* KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
* details.
* 
* You should have received a copy of the GNU General Public License along with
* KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA 02111-1307 USA
* 
* Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>.
*)

(** Compiled input clauses. *)

(** Notes:

   The compiled clauses intern references to predicate objects (i.e.
   objects of type Predicate.predicate) and information obtained by
   them. Abolishing a predicate can make a compiled clause unusable.

   See also notes for module [IcSet].
*)

type sign =
    Pos (** Positive. *)
  | StratifiedNegAsFailure (** Stratified negation as failure. 

                 [\+(A)], where [A] must be a non-builtin atom.
		 Succeeds iff [A] is not unifiable with a fact,
		 at time of the call. [A] is handled like a builtin
		 that has to be ground in subgoal sorting. 

		 This kind of negation should work well with
		 stratification and ground facts. More complex
		 expressions within the scope of [\+] (conjunctions,
		 existential variables) can be translated into this
		 form. *)
  | StratifiedNegAsFailureNotSubsumed 
    (** See explanation for 'not_subsumed' in user documentation. *)
  | Findall
  | Neg2

module BodyLiteral :
  sig

    type literal

    val get_sign : literal -> sign

    val get_predicate : literal -> Predicate.predicate

    val get_atom : literal -> Term.term

    val get_reader : literal -> Access.Reader.reader

    val get_foreign_context : literal -> Term.context

    val get_is_delta : literal -> bool

    val get_builtin : literal -> Term.p

    val make_literal : Predicate.predicate_table ->
      Term.symbol_table ->
      Term.term -> literal

    val print_optim_info : literal -> unit

  end


module HeadLiteral :
    sig

      type literal
	    
      val get_atom : literal -> Term.term

      val get_predicate : literal -> Predicate.predicate

      val make_literal : Predicate.predicate_table -> Term.term -> literal

    end

module Aux :
    sig

      (** Used to hold auxiliary datastructures. *)
      type aux

      (** Creates an [aux] for the given [Cfg.cfg].*)
      val make_aux : Cfg.cfg -> 
	Predicate.predicate_table -> Term.symbol_table ->
	  NegativeUnits.store -> aux

    end

module DisjunctiveHead :
    sig
      type head
      val get_literals : head -> HeadLiteral.literal list
    end

type body =
    Body of BodyLiteral.literal list
	
type head =
    False
  | Unit of HeadLiteral.literal
  | Conjunction of HeadLiteral.literal list
  | Disjunction of DisjunctiveHead.head

type clause

val get_body : clause -> body

val get_head : clause -> head

val get_news : clause -> News.news

val get_id : clause -> string

val get_source : clause -> Term.term

val make_clause : Aux.aux -> Term.term -> clause

(** Returns a list of clauses, that contains for each body literal
   with a predicate that appears in delta_predicates, a copy of the
   [clause], with the body literal replaced by ist delta version. 
*)
val clause_delta_clauses : Predicate.predicate list -> clause -> clause list

(** Prepare a clause for [eval_clause]. *)
val initialize_clause : clause -> News.news -> unit

(** Raises [BranchClosed]. *)
val eval_clause : clause -> unit

val reset_gen_id : unit -> unit

(** Raised if [False] head has been derived (a hypertableau branch has been
 closed. 
*)
exception BranchClosed of Fact.proof

