(*
* Copyright (C) 2003 Universitaet Koblenz-Landau
* 
* This file is part of KRHyper.
* 
* KRHyper is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
* 
* KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
* details.
* 
* You should have received a copy of the GNU General Public License along with
* KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA 02111-1307 USA
* 
* Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>.
*)

(** Storing and processing of newly derived facts (similar to a SOS). *)

(* # TODO ... bulk/single implementations, fact types, disjunction *)

(** The type of [news] objects. *)
type news

(** Create a [news] object. *)
val make_news : Cfg.cfg -> news

(** An unusable [news] object to initialize typed fields. *)
val null_news : news

val set_max_literal_weight : news -> int -> unit

val reset_weight_limit_exceeded : news -> unit

val unit_table_size : news -> int

val delayed_clauses_size : news -> int 

(** Is > 0 iff an inferred clause has been discarded since it exceeded
  the weight limit. The returned amount is the difference from the weight of
  the lightest such discarded clause to the limit.
*)
val weight_limit_exceeded : news -> int

(** Insert a fact into a [news] object. *)
val insert_unit : news -> Predicate.predicate -> Fact.fact -> unit

(** Insert a disjunction into a [news] object. The disjunction is
    represented as a list of length > 1 of quadruples. A quadruple
    contains: the predicate, the atom, the weight of the atom, and
    a list of terms used for complement splitting.
  *)       
val insert_disjunction : news -> 
  (Predicate.predicate * Fact.fact * int * (Fact.fact list)) list -> unit

(** [process_news news] processes facts in [news]. Processed facts
   are removed from [news]. 

   The number of actually new facts is returned, i.e. processed facts
   that are not subsumed by previously derived facts.

   The precise meaning of [process_news]  depends on the particular strategy.

   - For seminaive delta iteration, all facts in news are removed and, if
     they are not redundant by subsumption, are asserted as facts and
     as delta facts.
*)
val process_news : news -> int


(** Remove all unit facts from news. To be called in place of
   process_news when the facts should be thrown away (e.g. after
   branch closing) , but the news object should be used again (e.g. for
   alternative branches). *) 
val clear_news : news -> unit

module Disjunct :
  sig
    type disjunct
    val is_wp_semantics_true : disjunct -> bool
    val get_fact : disjunct -> Fact.fact
    val get_predicate : disjunct -> Predicate.predicate
    val get_complements : disjunct -> Fact.fact list
  end

type delayed_clause

val get_disjuncts : delayed_clause -> Disjunct.disjunct list

(** Raises [Not_found] if no disjunction that is not subsumed by a unit in
   the extent is found. 
 *)
val pick_disjunction : news -> delayed_clause

type state

val obtain_state : news -> state

val install_state : state -> unit
