(*
* Copyright (C) 2003 Universitaet Koblenz-Landau
* 
* This file is part of KRHyper.
* 
* KRHyper is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
* 
* KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
* details.
* 
* You should have received a copy of the GNU General Public License along with
* KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA 02111-1307 USA
* 
* Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>.
*)

(* let print_debug x = print_string "DEBUG: "; print_endline x *)

let print_debug x = ()
 
(****************************************************************************)
(* State *)
(****************************************************************************)

module State =
  struct

    type state =
	{ extent_state:Access.Store.state list;
	  negative_units_state:NegativeUnits.state
	}

    let obtain_extent_state predicate_table =
      let states = ref [] in
      Predicate.PredicateTable.iter
	(fun psymbol predicate ->
	  List.iter 
	    (fun writer -> 
	      states := 
		Access.Store.obtain_state (Access.Writer.get_store writer) ::
		!states)
	    (Predicate.get_writers predicate))
	predicate_table;
      !states

    let obtain_state predicate_table negative_units =
      { extent_state = obtain_extent_state predicate_table;
	negative_units_state = NegativeUnits.obtain_state negative_units }

    let add_layer state =
      { extent_state = List.map Access.Store.add_layer state.extent_state;
	negative_units_state =
	NegativeUnits.add_layer state.negative_units_state }
	
    let install_state state =
      List.iter Access.Store.install_state state.extent_state;
      NegativeUnits.install_state state.negative_units_state

  end


(****************************************************************************)
(* Refutation *)
(****************************************************************************)

type refutation =
    NullRefutation
  | PrimitiveRefutation of Fact.proof
  | ComposedRefutation of (refutation list)


(****************************************************************************)
(* Aux *)
(****************************************************************************)

module Aux =
  struct
    type aux = {
	predicate_table:Predicate.predicate_table;
	symbol_table:Term.symbol_table;
	cfg:Cfg.cfg;
	news:News.news;
	negative_units:NegativeUnits.store;
	input_state:State.state;
	use_proof_terms:bool;
	use_negative_units:bool;
	use_level_cut:bool;
	mutable proofs:Fact.proof list; (* deprecated *)
	mutable refutation:refutation;
	mutable stat_disjunction_nodes:int;
	mutable stat_stretchings:int;
	mutable stat_closed_branches:int;
	mutable stat_models:int;
	mutable stat_eval_rounds:int;
	mutable t_start:float;
	mutable t_model_start:float;
	mutable t_round_start:float;
	mutable n_strata:int;
	mutable stat_max_weight:int;
	mutable stat_depth:int;
	mutable stat_level_cuts:int;
      }

    let make_aux cfg predicate_table symbol_table negunits =
      let news = News.make_news cfg in
      let input_state =
	State.obtain_state predicate_table negunits in
      { predicate_table = predicate_table;
	symbol_table = symbol_table;
	cfg = cfg;
	news = news;
	negative_units = negunits;
	input_state = input_state;
	use_proof_terms = Cfg.get_proof_terms_flag cfg;
	use_negative_units = Cfg.get_negative_units_flag cfg;
	use_level_cut =
	Cfg.get_proof_terms_flag cfg &&	Cfg.get_level_cut_flag cfg;
	proofs = [];
	refutation = NullRefutation;
	stat_disjunction_nodes = 0;
	stat_stretchings = 0;
	stat_closed_branches = 0;
	stat_models = 0;
	stat_eval_rounds = 0;
	t_start=0.0;
	t_round_start=0.0;
	t_model_start=0.0;
	n_strata=0;
	stat_max_weight=(-1);
	stat_depth = 0;
	stat_level_cuts = 0
      }

    let inc_stat_disjunction_nodes aux =
      aux.stat_disjunction_nodes <- aux.stat_disjunction_nodes + 1
      (*
      Msg.msg1 3 "                Disjunction Node (#%d)."
	aux.stat_disjunction_nodes
       *)

    let inc_stat_stretchings aux =
      aux.stat_stretchings <- aux.stat_stretchings + 1;
      Msg.msg2 3 "                Resetting max weight to %d (#%d)."
	aux.stat_max_weight
	aux.stat_stretchings

    let inc_stat_closed_branches aux =
      aux.stat_closed_branches <- aux.stat_closed_branches + 1
      (*	  
      Msg.msg1 3 "                Branch closed (#%d)."
	aux.stat_closed_branches *)

    let inc_stat_models aux =
      aux.stat_models <- aux.stat_models + 1

    let inc_stat_eval_rounds aux =
      aux.stat_eval_rounds <- aux.stat_eval_rounds + 1

    let inc_stat_level_cuts aux =
      aux.stat_level_cuts <- aux.stat_level_cuts + 1;
      Msg.msg1 3 "                Level cut (#%d)."
	aux.stat_level_cuts


    let info_msg aux =
      let t = Sys.time () in
      Msg.msg1 2 "Model number:       %9d." aux.stat_models;
      Msg.msg1 2 "Time:               %8.3fs." (t -. aux.t_start);
      Msg.msg1 2 "Time in last model: %8.3fs." (t -. aux.t_model_start);
      Msg.msg1 2 "Disjunction nodes:  %9d." aux.stat_disjunction_nodes;
      Msg.msg1 2 "Level cuts:         %9d." aux.stat_level_cuts;
      Msg.msg1 2 "Tableau depth:      %9d." aux.stat_depth;
      Msg.msg1 2 "Closed branches:    %9d." aux.stat_closed_branches;
      Msg.msg1 2 "Evaluation rounds:  %9d." aux.stat_eval_rounds;
      Msg.msg1 2 "Max weight resets:  %9d." aux.stat_stretchings;
      Msg.msg1 2 "Latest max weight:  %9d." aux.stat_max_weight
  end


(****************************************************************************)
(* Stratum *)
(****************************************************************************)

module Stratum =
  struct

    type stratum = {
	plain_clauses:IcSet.clauses;
	inc_clauses:IcSet.clauses;
	delta_predicates:Predicate.predicate list;
	mutable needs_compilation:bool;
	aux:Aux.aux
      }

    let make_stratum aux plain_clauses =
      let plain_clauses =
	if aux.Aux.use_negative_units then
	  IcSet.negative_units_to_store aux.Aux.negative_units plain_clauses
	else
	  plain_clauses in
      let delta_predicates = (IcSet.idb_predicates plain_clauses) in
      { plain_clauses=IcSet.sort_clauses plain_clauses;
	inc_clauses=IcSet.sort_clauses
	  (IcSet.make_eval_inc_clauses plain_clauses delta_predicates);
	delta_predicates=delta_predicates;
	needs_compilation=true;
	aux=aux
      }

    let print_fat_line () =
      print_string "============================";
      print_endline "=================================================="
    
    let print_task_debug_info
	predicate_table
	delta_predicates
	plain_clauses
	inc_clauses =
      print_fat_line ();
      print_endline "COMPILATION INFORMATION: BEGIN";
      print_fat_line ();
      print_endline "COMPILATION INFORMATION: PREDICATES";
      print_fat_line ();
      Info.print_predicates_info predicate_table;
      print_fat_line ();
      print_endline "COMPILATION INFORMATION: PLAIN CLAUSES";
      print_fat_line ();
      Info.print_clauses_info plain_clauses;
      print_fat_line ();
      print_endline "COMPILATION INFORMATION: INC CLAUSES";
      print_fat_line ();
      Info.print_clauses_info inc_clauses;
      print_fat_line ();
      print_endline "COMPILATION INFORMATION: END";
      print_fat_line ()
	
   (* This can be called at any time to re-compile a clause set before
      evaluation. Actual extent sizes then are considered for 
      join ordering. It must be called at least once before evaluation. *)
    let ensure_compiled stratum =
      if stratum.needs_compilation then
	let aux = stratum.aux in
	let news = aux.Aux.news in
	Msg.msg1 2 "Compiling stratum (%d rules)."
	  (List.length stratum.plain_clauses);
	List.iter (fun clause -> Ic.initialize_clause clause news)
	  stratum.plain_clauses;
	List.iter (fun clause -> Ic.initialize_clause clause news)
	  stratum.inc_clauses;
	stratum.needs_compilation <- false;
	Msg.msg 2 "Compilation of stratum finished.";
	if Cfg.get_debug_flag aux.Aux.cfg then
	  print_task_debug_info 
	    aux.Aux.predicate_table
	    stratum.delta_predicates
	    stratum.plain_clauses 
	    stratum.inc_clauses	    
	    
    let print_compiled_clauses stratum =
      ensure_compiled stratum;
      Info.print_clauses_as_terms stratum.plain_clauses

  end


(****************************************************************************)
(* Node *)
(****************************************************************************)

type literal = 
    Literal of News.Disjunct.disjunct
  | UnboundLiteral

type node =
    { strata:Stratum.stratum list;
      aux:Aux.aux;
      mutable max_weight:int;
      parent:node;
      mutable state:State.state;
      mutable initial_news_state:News.state;
      literal: literal;
      right_brother_literals:News.Disjunct.disjunct list;
      mutable stretch_ancestor:stretch_ancestor;
      mutable left_brother_refutations: refutation list;
      mutable refutation: refutation;
      depth:int
    }


(****************************************************************************)
(* Stretch Ancestor *)

and stretch_ancestor =
    StretchAncestor of node * int (* node * weight_increment *)
  | UnboundStretchAncestor

let make_stretch_ancestor node weight_increment =
  StretchAncestor (node, weight_increment)

(****************************************************************************)
(* Weight Management Parameters *)

let wm_initial_max_weight cfg =
  Cfg.get_max_weight_initial_param cfg

let wm_initial_max_weight_next_stratum node =
    node.max_weight

let wm_child_max_weight n = n

let wm_increase_max_weight_of_stretch_root aux node required_increment =
  let max_weight =
    node.max_weight + 
      (max 
	 (Cfg.get_max_weight_increment_param aux.Aux.cfg)
	 required_increment) in
  node.max_weight <- max_weight;
  aux.Aux.stat_max_weight <- max_weight


(****************************************************************************)
(* Node Construction and Access *)

let make_initial_node clauses aux =
  let cfg = aux.Aux.cfg 
  and news = aux.Aux.news in
  let strata =
    List.map 
      (fun clauses1 -> Stratum.make_stratum aux clauses1) 
      (IcSet.stratify clauses) in
  aux.Aux.n_strata <- (List.length strata);
  Msg.msg1 2 "Number of strata: %5d" (List.length strata);
  let rec node =
    { strata = strata;
      aux = aux;
      max_weight = wm_initial_max_weight cfg;
      parent = node; (* self for "unbound" *)
      state = 
      if Cfg.get_merge_inf_input_flag cfg then
	aux.Aux.input_state
      else
	State.add_layer aux.Aux.input_state;
      initial_news_state = News.obtain_state news;
      literal = UnboundLiteral;
      right_brother_literals = [];
      stretch_ancestor = UnboundStretchAncestor;
      left_brother_refutations = [];
      refutation = NullRefutation;
      depth = 0 } in
  node

let has_next_stratum node =
  match node.strata with
    _ :: _ :: _ -> true
  | [_] | [] -> false

let make_next_stratum_node node news_state =
  let strata =
    match node.strata with
      _ :: strata1 -> strata1 
    | _ -> raise (Failure "No next stratum.");
  in
  { strata = strata;
    aux = node.aux;
    max_weight = wm_initial_max_weight_next_stratum node;
    parent = node; (* for backtracking (find_next_node) *)
    state = node.state; (* continue with state passed from lower stratum *)
    initial_news_state = news_state;
    literal = UnboundLiteral;
    right_brother_literals = [];
    stretch_ancestor = UnboundStretchAncestor;
    left_brother_refutations = [];
    refutation = NullRefutation;
    depth = node.depth + 1 }

let make_child_node node lit lits news_state =
  let aux = node.aux in
  aux.Aux.stat_depth <- max aux.Aux.stat_depth (node.depth + 1);
  { strata = node.strata;
    aux = aux;
    max_weight = wm_child_max_weight node.max_weight;
    parent = node;
    state = State.add_layer node.state;
    initial_news_state = news_state;
    literal = Literal lit;
    right_brother_literals = lits;
    stretch_ancestor = node.stretch_ancestor;
    left_brother_refutations = [];
    refutation = NullRefutation;
    depth = node.depth + 1;
  }

let make_right_brother_node node lit lits =
  { strata = node.strata;
    aux = node.aux;
    max_weight = node.max_weight;
    parent = node.parent;
    state = State.add_layer node.parent.state;
    initial_news_state = node.initial_news_state;
    literal = Literal lit;
    right_brother_literals = lits;
    stretch_ancestor = node.stretch_ancestor;
    left_brother_refutations =
    node.refutation :: node.left_brother_refutations;
    refutation = NullRefutation;
    depth = node.depth }

(****************************************************************************)
(* For Debugging *)
(*
let debug_print_node node =
  match node.literal with
    Literal disjunct ->
      PrologPrint.print
	(Fact.get_term (News.Disjunct.get_fact disjunct))
  | UnboundLiteral ->
      print_string "[UnboundLiteral]"
*)
(****************************************************************************)

let insert_disjunct aux disjunct =
  if Cfg.get_wp_semantics_flag aux.Aux.cfg && 
    News.Disjunct.is_wp_semantics_true disjunct then
    ()
  else
    let fact = News.Disjunct.get_fact disjunct in
    let term = Fact.get_term fact
    and predicate = News.Disjunct.get_predicate disjunct in

    if Cfg.get_show_derived_literals_flag aux.Aux.cfg then
      begin
	print_string "selected_disjunct(";
	PrologPrint.print term;
	print_endline ", _)."
      end;

    Predicate.insert predicate term fact;
    if Predicate.has_delta_predicate predicate then
      Predicate.insert
	(Predicate.get_delta_predicate predicate) term fact;
    if aux.Aux.use_negative_units then
      let negunits = aux.Aux.negative_units in
      List.iter 
	(fun complement ->
	  NegativeUnits.insert negunits complement)
	(News.Disjunct.get_complements disjunct)


(****************************************************************************)
(* Eval *)


type eval_result =
    Open of int * int (* of weight_limit_exceeded * delta_count *)
  | Closed of Fact.proof

let rec eval_1 
    aux news delta_predicates weight_limit_exceeded clauses clauses1 =
  Aux.inc_stat_eval_rounds aux;
  aux.Aux.t_round_start <- (Sys.time ());
  let result = 
    try  
      List.iter Ic.eval_clause clauses;
      let weight_limit_exceeded_1 =
	if weight_limit_exceeded = 0 then
	  News.weight_limit_exceeded news
	else
	  weight_limit_exceeded in
      List.iter Predicate.clear_delta_predicate delta_predicates;
      Open (weight_limit_exceeded_1, (News.process_news news))
    with Ic.BranchClosed proof ->
      List.iter Predicate.clear_delta_predicate delta_predicates;
      News.clear_news news;
      Closed proof in
  match result with
    Open (wle, delta_count) ->
      Msg.msg4 2 "Round: %9d, Delta: %9d, Delayed: %9d, Time: %7.2fs"
	aux.Aux.stat_eval_rounds
	delta_count
	(News.delayed_clauses_size news)
	(Sys.time () -. aux.Aux.t_round_start);
      if delta_count > 0 then
	eval_1 
	  aux news delta_predicates wle clauses1 clauses1
      else
	result
  | Closed proof ->
      result

let neg_proof fact neg_fact =
  match Fact.get_proof fact with
    Fact.BuiltinProof ->
      Fact.BuiltinProof
  | Fact.DisjProof _ ->
      Fact.NegProof
	(Term.Unbound,
	 [|fact|],
	 neg_fact)
  | _ ->
      raise (Failure "Unexpected proof type for tableau node literal.")
	
let eval node =
  print_debug "EVAL";
  let aux = node.aux in
  let news = aux.Aux.news in
  State.install_state node.state;
  News.install_state node.initial_news_state;
  News.set_max_literal_weight news node.max_weight;
  News.reset_weight_limit_exceeded news;
  try
    if aux.Aux.use_negative_units then
      match node.literal with
	Literal lit ->
	  let fact = News.Disjunct.get_fact lit in
	  let neg_fact = 
	    NegativeUnits.lookup node.aux.Aux.negative_units fact in
	  Closed
	    (if aux.Aux.use_proof_terms then
	      begin
		print_debug "EVAL-CLOSED-BY-COMPLEMENT";
		neg_proof fact neg_fact
	      end
	    else
	      Fact.BuiltinProof)
      | UnboundLiteral ->
	  raise Not_found
    else
      raise Not_found
  with Not_found ->
    match node.strata with
      stratum :: _ ->
	begin
	  Stratum.ensure_compiled stratum;
	  match node.literal with
	    Literal literal ->
	      (* DISJUNCTION NODE *)
	      insert_disjunct aux literal;

	      eval_1
		aux
		news
		stratum.Stratum.delta_predicates
		0
		stratum.Stratum.inc_clauses
		stratum.Stratum.inc_clauses
	  | UnboundLiteral ->
	      (* FIRST NODE OR FIRST NODE OF A HIGHER STRATUM *)
	      eval_1
		aux
		news
		stratum.Stratum.delta_predicates
		0
		stratum.Stratum.plain_clauses
		stratum.Stratum.inc_clauses
	end
    | [] ->
	Msg.msg 2 "No stratum.";
	Open (0,0)
	  
(****************************************************************************)
(* Find Model Loop *)
(****************************************************************************)

type result =
    Model of node
  | NoModel of Aux.aux

type disjunction =
    Disjunction of News.delayed_clause
  | UnboundDisjunction

type next_node =
    NextNode of node
  | NoNextNode of node (* returns the top node *)

let info_find_model node =
  let aux = node.aux in
  let t = Sys.time () in
  Msg.msg4 2
    "Depth: %d, Weight: %d, Stratum: %d, Model:%7.2fs" 
    (* (t -. aux.Aux.t_start) *)
    node.depth
    node.max_weight
    (aux.Aux.n_strata - (List.length node.strata))
    (t -. aux.Aux.t_model_start)

let install_top_state aux =
  State.install_state aux.Aux.input_state

(* ### Use level (depth) of Closed as parameter to level_cut ! *)
(* but a refutation propagated upwards may use a different depth... *)

(** Raises Not_found if proof is Fact.UnknownProof. *)
let rec proof_iter_facts f proof =
  match proof with
    Fact.BuiltinProof | Fact.InputFactProof ->
      ()
  | Fact.Proof (_, facts) ->
      Array.iter 
	(fun fact -> f fact; proof_iter_facts f (Fact.get_proof fact))
	facts
  | Fact.DisjProof (_, _, facts, neg_facts) ->
      Array.iter 
	(fun fact -> f fact; proof_iter_facts f (Fact.get_proof fact))
	facts;
      List.iter 
	(fun fact -> f fact; proof_iter_facts f (Fact.get_proof fact))
	neg_facts
  | Fact.NegProof (_, facts, neg_fact) ->
      Array.iter 
	(fun fact -> f fact; proof_iter_facts f (Fact.get_proof fact))
	facts;
      f neg_fact;
      proof_iter_facts f (Fact.get_proof neg_fact)
  | Fact.UnknownProof ->
      raise Not_found

let rec refutation_iter_facts f refutation =
  match refutation with
    NullRefutation ->
      ()
  | PrimitiveRefutation proof ->
      proof_iter_facts f proof
  | ComposedRefutation refutations ->
      List.iter (fun r -> refutation_iter_facts f r) refutations

exception Found

let used_in_refutation facts refutation depth =
  (* Depth could be used to restrict the set of facts to iterate over. *)
  try
    refutation_iter_facts
      (fun fact ->
	if List.mem fact facts then raise Found) 
        (* ### Fact-identity: = (and not ==) *)
      refutation;
    false
  with Found | Not_found ->
    true

let level_cut node depth =
  if  depth >= 0 && node.aux.Aux.use_level_cut then
    match node.literal with
      UnboundLiteral ->
	(* Top node or top node of stratum. *)
	false
    | Literal lit ->
	(* Disjunction node. *)
	let facts = 
	  (News.Disjunct.get_fact lit) ::
	  (News.Disjunct.get_complements lit) in
	not (used_in_refutation facts node.refutation depth) 
  else
    false

let rec find_next_node node depth =
  if level_cut node depth then
    begin
      Aux.inc_stat_level_cuts node.aux;
      node.parent.refutation <- node.refutation;
      find_next_node
	node.parent 
	(if depth >= 0 then (depth + 1) else depth)
    end
  else
    match node.right_brother_literals with
      [] ->
	begin
	  if node.parent == node then
	    (* Top node. *)
	    begin
	      install_top_state node.aux;
	      NoNextNode node
	    end
	  else
	    match node.literal with
	      Literal _ ->
		(* Disjunction branch node. *)
		print_debug "NN_PARENT";
		node.parent.refutation <-
		  ComposedRefutation 
		    (node.refutation :: node.left_brother_refutations);
		find_next_node
		  node.parent 
		  (if depth >= 0 then (depth + 1) else depth)
	    | UnboundLiteral ->
		(* Top node of stratum, the parent is on a lower stratum *)
		node.parent.refutation <-
		  ComposedRefutation 
		    (node.refutation :: node.left_brother_refutations);
		find_next_node node.parent (-1)
	end
    | lit :: lits ->
	print_debug "NN_BROTHER";
	NextNode (make_right_brother_node node lit lits)
	  
let rec find_model node =
  info_find_model node;
  let aux = node.aux in
  let news = aux.Aux.news in
  match eval node with
    Open (weight_limit_exceeded, _) ->
      (* TABLEAU BELOW NODE IS OPEN *)
      (match node.stretch_ancestor with
	UnboundStretchAncestor ->
	  if weight_limit_exceeded > 0 then
	    if (Cfg.get_max_weight_increment_param aux.Aux.cfg) >= 0 then
	      (* SET STRETCH ANCESTOR *)
	      begin
		Msg.msg 3 "Weight limit exceeded.";
		print_debug "SET STRETCH ANCESTOR";
		node.stretch_ancestor <-
		  StretchAncestor (node, weight_limit_exceeded)
	      end
	    else
	      Msg.msg 3 "Weight limit exceeded: model may be incomplete.";
      | StretchAncestor _ ->
	  ());
      (* TRY TO EXTEND HYPERTABLEAU *)
      (match 
	try
	  Disjunction (News.pick_disjunction news)
	with Not_found ->
	  UnboundDisjunction with
	Disjunction disjunction ->
	  print_debug("EXTEND");
	  Aux.inc_stat_disjunction_nodes aux;
	  print_debug "CHILD";
	  (* PICKED A DISJUNCTION FROM SOS *)
	  (match (News.get_disjuncts disjunction) with
	    lit :: lits ->
	      let child =
		make_child_node node lit lits (News.obtain_state news)
	      in
	      if Cfg.get_show_derived_literals_flag aux.Aux.cfg then
		print_endline "attached_disjunction(_).";
	      find_model child
	  | [] ->
	      raise (Failure "Empty disjunction."))
      | UnboundDisjunction ->
	  (* SOS EMPTY *)
	  match node.stretch_ancestor with
	    UnboundStretchAncestor ->
	      print_debug("UNBOUND SA");

	      (* MODEL FOR THIS STRATUM FOUND *)
	      if has_next_stratum node then
		(* NEXT STRATUM *)
		begin
		  print_debug "NEXT STRATUM";
		  find_model 
		    (make_next_stratum_node 
		       node 
		       (News.obtain_state news)) (* should be empty here *)
		end
	      else
		(* MODEL FOR ALL STRATA FOUND *)
		begin
		  Aux.inc_stat_models aux;
		  Model node
		end
	  | StretchAncestor (snode, weight_increment) ->
	      (* REDO FROM ANCESTOR WITH INCREASED MAX-WEIGHT *)
	      print_debug("STRETCH");
	      wm_increase_max_weight_of_stretch_root
		aux
		snode
		weight_increment;
	      Aux.inc_stat_stretchings aux;
	      snode.stretch_ancestor <- UnboundStretchAncestor;
	      match snode.literal with
		UnboundLiteral ->
		  (* FIRST NODE OR FIRST NODE OF A HIGHER STRATUM *)
		  (* RE-USE THE NODE'S STATE AS LEMMAS *)
		  print_debug("STRETCH FIRST");
		  find_model snode
	      | Literal lit -> 
		  print_debug("STRETCH DISJ");
		  snode.state <- State.add_layer snode.parent.state;
		  find_model snode)
  | Closed proof ->
      Aux.inc_stat_closed_branches aux;
      if aux.Aux.use_proof_terms then
	begin
	  aux.Aux.proofs <- proof :: aux.Aux.proofs;
	  node.refutation <- PrimitiveRefutation proof
	end;
      match find_next_node node 0 with
	NextNode next_node ->
	  find_model next_node
      | NoNextNode top_node ->
	  aux.Aux.refutation <- top_node.refutation;
	  NoModel aux


(****************************************************************************)
(* Invocation *)
(****************************************************************************)

let result_aux result =
  match result with
    Model node -> node.aux 
  | NoModel aux -> aux

let is_open result =
  match result with
    Model _ -> true
  | NoModel _ -> false

let info_msg result =
  Aux.info_msg (result_aux result)

let initial_result cfg predicate_table symbol_table negunits =
  let aux = Aux.make_aux cfg predicate_table symbol_table negunits in
  NoModel aux

let first_result cfg predicate_table symbol_table negunits clauses =
  NegativeUnits.clear negunits;
  let aux = Aux.make_aux cfg predicate_table symbol_table negunits in
  let t = Sys.time () in
  aux.Aux.t_start <- t;
  aux.Aux.t_model_start <- t;
  let node = (make_initial_node clauses aux) in
  find_model node
    
let abandon_result result =
  if is_open result then
    let aux = (result_aux result) in
    install_top_state aux;
    NoModel aux
  else
    let aux = (result_aux result) in
    install_top_state aux;
    result

let next_result result =
  match result with
    Model node ->
      node.aux.Aux.t_model_start <-  Sys.time ();
      (match find_next_node node (-1) with
	NextNode next_node ->
	  find_model next_node
      | NoNextNode top_node ->
	  node.aux.Aux.refutation <- top_node.refutation;
	  NoModel node.aux)
  | NoModel _ ->
      result

let get_l_proof result =
  (result_aux result).Aux.proofs

let get_refutation result =
  (result_aux result).Aux.refutation

(****************************************************************************)

let gensym_count = ref 0

let gensym string =
  gensym_count := !gensym_count+1;
  string^string_of_int !gensym_count


let rec rtt st aux r =
  match r with
    NullRefutation ->
      Fact.proof_term_for_r_proof st aux Fact.BuiltinProof
  | PrimitiveRefutation proof ->
      Fact.proof_term_for_r_proof st aux proof
  | ComposedRefutation rs ->
      Convert.array_to_compound
	st
	"node"
	[|Convert.string_to_term st (gensym "composed");
	  Convert.string_to_term st "---";
	  Convert.string_to_term st "composed";
	  (Convert.list_of_term_to_term
	     st
	     (List.map (fun r1 -> rtt st aux r1) rs));
	  Convert.list_of_term_to_term st []|]

let refutation_to_term symbol_table r =
  let aux = Fact.make_proof_term_aux () in
  rtt symbol_table aux r

(****************************************************************************)


let print_compiled_rules cfg predicate_table symbol_table negunits clauses =
  (* Mimic some steps performed at first_result *)
  NegativeUnits.clear negunits;
  let aux = Aux.make_aux cfg predicate_table symbol_table negunits in
  let t = Sys.time () in
  aux.Aux.t_start <- t;
  aux.Aux.t_model_start <- t;
  let cfg = aux.Aux.cfg 
  and news = aux.Aux.news in
  let strata =
    List.map 
      (fun clauses1 -> Stratum.make_stratum aux clauses1) 
      (IcSet.stratify clauses)
  in
  aux.Aux.n_strata <- (List.length strata);
  let n = ref 0 in
  print_string "% ========== Number of strata: ";
  print_int (List.length strata);
  print_endline " ==========";
  List.iter
    (fun stratum ->
      print_string "% ========== Stratum: ";
      print_int !n;
      print_endline " ==========";
      Stratum.print_compiled_clauses stratum;
      n := !n + 1)
    strata



